class Osc{
    constructor() {
        this.type = "sawtooth";
        this.volume = -10;
        this.detune = 0;
        this.phase = 0;
    }
}

class Env{
    constructor() {
        this.attack = 0.05;
        this.decay = 0.05;
        this.sustain = 1;
        this.release = 0.5;
        this.attackCurve = "exponential";
        this.decayCurve = "linear";
        this.releaseCurve = "exponential";
    }
}

const waveform1Select = document.getElementById('waveform-1');
const waveform2Select = document.getElementById('waveform-2');
const volume1Slider = document.getElementById('volume-1');
const volume2Slider = document.getElementById('volume-2');
const detune1Slider = document.getElementById('detune-1');
const detune2Slider = document.getElementById('detune-2');
const phase1Slider = document.getElementById('phase-1');
const phase2Slider = document.getElementById('phase-2');

const attackSlider = document.getElementById("attack");
const decaySlider = document.getElementById("decay");
const sustainSlider = document.getElementById("sustain");
const releaseSlider = document.getElementById("release");

const filterTypeSelect = document.getElementById("filter-type");
const filterRolloffSelect = document.getElementById("filter-rolloff");
const filterFrequencySlider = document.getElementById("filter-frequency");
const filterQSlider = document.getElementById("filter-q");

const frequencyInput = document.getElementById('frequency');
const playPauseButton = document.getElementById('play-pause');
const keyboard = document.querySelector(".keyboard");

const synthControls = document.getElementById("synth-controls");
const sliders = synthControls.querySelectorAll("input[type='range']");

sliders.forEach((slider) => {
    if (slider.id === "filter-frequency") return;

    const sliderValue = document.createElement("div");
    sliderValue.classList.add("slider-value");
    sliderValue.id = `${slider.id}-value`;

    slider.addEventListener("input", function () {
        sliderValue.textContent = slider.value;
    });

    slider.addEventListener("dblclick", function () {
        slider.value = slider.getAttribute("value");
        sliderValue.textContent = slider.value;
        slider.dispatchEvent(new Event("input"));
    });

    slider.parentNode.insertBefore(sliderValue, slider.nextSibling);
});

waveform1Select.addEventListener('change', () => {
    document.getElementById(oscillator1.type + "-pic1").style.display = "none";

    oscillator1.type = waveform1Select.value
    synth1.oscillator.type = waveform1Select.value;

    document.getElementById(waveform1Select.value + "-pic1").style.display = "inline-block"
});

waveform2Select.addEventListener('change', () => {
    document.getElementById(oscillator2.type + "-pic2").style.display = "none";

    oscillator2.type = waveform2Select.value;
    synth2.oscillator.type = waveform2Select.value;

    document.getElementById(waveform2Select.value + "-pic2").style.display = "inline-block"
});

volume1Slider.addEventListener('input', () => {
    oscillator1.volume = volume1Slider.value;
    synth1.oscillator.volume.value = volume1Slider.value;
});

detune1Slider.addEventListener('input', () => {
    oscillator1.detune = detune1Slider.value;
    synth1.detune.value = detune1Slider.value;
});

phase1Slider.addEventListener('input', () => {
    oscillator1.phase = phase1Slider.value;
    synth1.oscillator.phase.value = phase1Slider.value;
});

volume2Slider.addEventListener('input', () => {
    oscillator2.volume = volume2Slider.value;
    synth2.oscillator.volume.value = volume2Slider.value;
});

detune2Slider.addEventListener('input', () => {
    oscillator2.detune = detune2Slider.value;
    synth2.detune.value = detune2Slider.value;
});

phase2Slider.addEventListener('input', () => {
    oscillator2.phase = phase2Slider.value;
    synth2.oscillator.phase.value = phase2Slider.value;
});

attackSlider.addEventListener('input', () => {
    envelope.attack = attackSlider.value;
    synth1.envelope.attack = attackSlider.value;
    synth2.envelope.attack = attackSlider.value;
})

decaySlider.addEventListener('input', () => {
    envelope.decay = decaySlider.value;
    synth1.envelope.decay = decaySlider.value;
    synth2.envelope.decay = decaySlider.value;
})

sustainSlider.addEventListener('input', () => {
    envelope.sustain = sustainSlider.value;
    synth1.envelope.sustain = sustainSlider.value;
    synth2.envelope.sustain = sustainSlider.value;
})

releaseSlider.addEventListener('input', () => {
    envelope.release = releaseSlider.value;
    synth1.envelope.release = releaseSlider.value;
    synth2.envelope.release = releaseSlider.value;
})

filterTypeSelect.addEventListener('change', () => {
    filter.type = filterTypeSelect.value;
});

filterRolloffSelect.addEventListener('change', () => {
    filter.rolloff = filterRolloffSelect.value;
});

filterQSlider.addEventListener('input', () => {
    filter.Q.value = filterQSlider.value;
})

frequencyInput.addEventListener('input', () => {
    const frequency = parseFloat(frequencyInput.value);
    if (!isNaN(frequency)) {
        note = frequency;
        if (lastActiveNote) {
            lastActiveNote.classList.remove('active');
        }
    }
});

playPauseButton.addEventListener('mousedown', async () => {
    if (Tone.context.state !== 'running') {
        try {
            await Tone.start();
            console.log('Audio context started');
        } catch (error) {
            console.error('Failed to start audio context:', error);
        }
    }

    if (! isPlaying) {
        synth1.triggerAttack(note);
        synth2.triggerAttack(note);
        isPlaying = true;
    }
})

playPauseButton.addEventListener('mouseup', () => {
    synth1.triggerRelease();
    synth2.triggerRelease();
    isPlaying = false;
})

playPauseButton.addEventListener('mouseleave', () => {
    synth1.triggerRelease();
    synth2.triggerRelease();
    isPlaying = false;
})

document.addEventListener('keydown', (evt) => {
    if (evt.key === ' ') {
        if (! isPlaying) {
            synth1.triggerAttack(note);
            synth2.triggerAttack(note);
            isPlaying = true;
        }
    }
})

document.addEventListener('keyup', (evt) => {
    if (evt.key === ' ') {
        synth1.triggerRelease();
        synth2.triggerRelease();
        isPlaying = false;
    }
})

// Start

const oscillator1 = new Osc();
const oscillator2 = new Osc();
oscillator2.volume = -60;

let filter = new Tone.Filter({
    type: "lowpass",
    frequency: 20000,
    Q: 1,
    rolloff: -12,
}).toDestination();

let envelope = new Env();

let synth1 = new Tone.Synth({
    oscillator: oscillator1,
    envelope: envelope,
    volume: -10,
}).connect(filter);

let synth2 = new Tone.Synth({
    oscillator: oscillator2,
    envelope: envelope,
    volume: -10,
}).connect(filter);

let [points, transformFunc] = scaleTransform(19, 22000, [1, 10, 50, 100, 200, 500, 1000]);
filterFrequencySlider.max = points;
filterFrequencySlider.value = points;

const sliderValue = document.createElement("div");
sliderValue.classList.add("slider-value");
sliderValue.id = `filter-frequency-value`;

filterFrequencySlider.addEventListener("input", function () {
    let sliderVal = filterFrequencySlider.value;
    filter.frequency.value = transformFunc(sliderVal);
    sliderValue.textContent = parseInt(filter.frequency.value);
});

filterFrequencySlider.addEventListener("dblclick", function () {
    filterFrequencySlider.value = points    ;
    sliderValue.textContent = "22000";
});

filterFrequencySlider.parentNode.insertBefore(sliderValue, filterFrequencySlider.nextSibling);

let note = 440;
let lastActiveNote = null;
let isPlaying = false;

let noteFreq = createNoteTable();

noteFreq.forEach((keys, idx) => {
    const keyList = Object.entries(keys);
    const octaveElem = document.createElement("div");
    octaveElem.className = "octave";

    keyList.forEach((key) => {
        // if (key[0].length === 1) {
        octaveElem.appendChild(createKey(key[0], idx, key[1]));
        // }
    });

    keyboard.appendChild(octaveElem);
});

document
    .querySelector("div[data-note='F'][data-octave='5']")
    .scrollIntoView(false);


/**
 * Credit: https://developer.mozilla.org/en-US/docs/Web/API/Web_Audio_API/Simple_synth
 * Return an array of note frequency
 *
 * @return {[]} An array where [i]["C#"] is the frequency of C# on i'th octave
 */
function createNoteTable() {
    const noteFreq = [];
    for (let i=0; i< 8; i++) {
        noteFreq[i] = [];
    }

    const noteName = [
        "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"
    ]

    noteFreq[0]["A"] = 27.500000000000000;
    noteFreq[0]["A#"] = 29.135235094880619;
    noteFreq[0]["B"] = 30.867706328507756;

    noteFreq[1]["C"] = 32.703195662574829;
    noteFreq[1]["C#"] = 34.647828872109012;
    noteFreq[1]["D"] = 36.708095989675945;
    noteFreq[1]["D#"] = 38.890872965260113;
    noteFreq[1]["E"] = 41.203444614108741;
    noteFreq[1]["F"] = 43.653528929125485;
    noteFreq[1]["F#"] = 46.249302838954299;
    noteFreq[1]["G"] = 48.999429497718661;
    noteFreq[1]["G#"] = 51.913087197493142;
    noteFreq[1]["A"] = 55.000000000000000;
    noteFreq[1]["A#"] = 58.270470189761239;
    noteFreq[1]["B"] = 61.735412657015513;

    // Build the rest of the keys using frequency from one octave down
    for (let i = 2; i < 8; i += 1) {
        for (let note of noteName) {
            noteFreq[i][note] = 2 * noteFreq[i-1][note];
        }
    }

    return noteFreq;
}

/**
 * Credit: https://developer.mozilla.org/en-US/docs/Web/API/Web_Audio_API/Simple_synth
 * Return an element representing a note using
 *
 * @param {String} note Name of the note
 * @param {number} octave Octave it is on
 * @param {number} freq Frequency of the note
 * @return {HTMLDivElement}
 */
function createKey(note, octave, freq) {
    const keyElement = document.createElement("div");
    const labelElement = document.createElement("div");

    keyElement.className = "key";
    keyElement.dataset["octave"] = octave;
    keyElement.dataset["note"] = note;
    keyElement.dataset["frequency"] = freq;

    if (note.length > 1) {
        keyElement.classList.add("black");
    }

    labelElement.innerHTML = `${note}<sub>${octave}</sub>`;
    keyElement.appendChild(labelElement);

    keyElement.addEventListener("mousedown", notePressed, false);
    keyElement.addEventListener("mouseup", noteReleased, false);
    keyElement.addEventListener("mouseover", notePressed, false);
    keyElement.addEventListener("mouseleave", noteReleased, false);

    return keyElement;
}

/**
 * Credit: https://developer.mozilla.org/en-US/docs/Web/API/Web_Audio_API/Simple_synth
 * Handler for note mousedown/mouseover event
 *
 * @param event
 */
async function notePressed(event) {
    if (event.buttons % 2 === 1) {
        const dataset = event.target.dataset;

        if (!dataset["pressed"]) {
            if (Tone.context.state !== 'running') {
                try {
                    await Tone.start();
                    console.log('Audio context started');
                } catch (error) {
                    console.error('Failed to start audio context:', error);
                }
            }

            note = dataset["frequency"];
            synth1.triggerAttack(note);
            synth2.triggerAttack(note);
            frequencyInput.value = note;

            dataset["pressed"] = "yes";
        }
    }
}

/**
 * Credit: https://developer.mozilla.org/en-US/docs/Web/API/Web_Audio_API/Simple_synth
 * Handler for note mouseup/mouseleave event
 *
 * @param event
 */
function noteReleased(event) {
    const dataset = event.target.dataset;

    if (dataset && dataset["pressed"]) {
        synth1.triggerRelease();
        synth2.triggerRelease();

        delete dataset["pressed"];
    }
}



/**
 * Credit: https://stackoverflow.com/a/56557377
 * Calculate out transform function given the
 *
 * @param {number} min Minimum value we need to reach
 * @param {number} max Maximum value we need to reach
 * @param {[number]} intervals "Steps" at each interval
 * @return {[number,(function(number): number)]} The number of points on our slider and the transform function
 */
function scaleTransform(min, max, intervals){
    //determine how many "points" we need
    let distributions = intervals.length;
    let descretePoints = Math.ceil(
        (max - min) /
        intervals.reduce((total, step) => total + step / distributions, 0)
    );

    return [
        descretePoints,
        (input) => {
            let stepTransforms = intervals.map((s, i) => {
                let setCount = Math.min(
                    Math.ceil(input - (descretePoints * i) / distributions),
                    Math.round(descretePoints / distributions)
                );
                return setCount > 0 ? setCount * s : 0;
            });

            let lastStep = 0;
            let out =
                Math.round(
                    stepTransforms.reduce((total, num, i) => {
                        if (num) {
                            lastStep = i;
                        }
                        return total + num;
                    })
                ) + min;

            let currentUnit = intervals[lastStep];
            return Math.min(
                Math.round(out / currentUnit) * currentUnit, //round to nearest step
                max
            );
        },
    ];
}
